/*
 * Copyright (c) 1991, by Sun Microsystems Inc.
 */

#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/debug.h>
#include <sys/uio.h>
#include <sys/buf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/kmem.h>
#include <sys/poll.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/file.h>
#include <sys/sunddi.h>
#include "krtv.h"
#include "ramdisk.h"


#define	DEFAULT_MAXPHYS	(63 * 1024)			
#define TVTWO_SUPPORT
#define MAX_QUEUE_LEVEL 60
#define AUDIO_RAM_SIZE 512*1024
#define QUEUE_AUDIO_FRAME 0x101
#define I_SET_AUDIO_RATE 0x102

#define MAX_AUDIO_QUEUE 512

typedef struct {
	int frame_no;
	int size;
	int sync_clock;
	caddr_t addr;
	}AUDIO_FRAME;

typedef struct {
	int level;
	caddr_t base;
	int mem_offset;
	AUDIO_FRAME *head;
	AUDIO_FRAME *tail;
	AUDIO_FRAME *input;
	AUDIO_FRAME *output;
	AUDIO_FRAME audio_frame[MAX_AUDIO_QUEUE];
	}AUDIO_QUEUE_INFO;


/*
 * EXTERNAL DEPENDANCIES of TVTWO
SHM_FIELD_QUE *global_shm_que;
 */
char _depends_on[] = "drv/tvtwo";
extern SHM_FIELD_QUE *global_shm_que;

static u_int ramdisk_put_ram_intr();
static u_int ramdisk_get_ram_intr();

static int ramdisk_maxphys = DEFAULT_MAXPHYS;
static int base_offset[4]= {0,MAX_SHM_SIZE,MAX_SHM_SIZE*2,MAX_SHM_SIZE*3};
static int max_offset[4]= {MAX_SHM_SIZE,MAX_SHM_SIZE*2,MAX_SHM_SIZE*3,
		MAX_SHM_SIZE*4};
typedef struct{
	int wid;
	int start_x;
	int start_y;
	int width;
	int height;
	}START_INFO;

/*
 * The entire state of each ramdisk device.
 */
typedef struct {
	RAM_FIELD_QUE *ram_shm_que;
	int		ramsize;	/* how much is there */
	int		maxphys;	/* how big per transfer */
	ddi_iblock_cookie_t iblk_put_ram_cookie; /* for blocking intrs */
	ddi_iblock_cookie_t iblk_get_ram_cookie; /* for blocking intrs */
	ddi_softintr_t	put_ram_id;
	ddi_softintr_t	get_ram_id;
	kmutex_t put_ram_intr_mutex;
	kmutex_t get_ram_intr_mutex;
	int 		blkno;
	int		interlace;
	int		rtv_open;
	int		que_size_in;
	int		que_size_out;
	int		user_mapped;
	int		stream_id;
	int		start_play;
	int		start_record;
	int		blocking_read;
	int		blocking_write;
	struct	buf	userbuf;
	dev_info_t	*dip;		/* my devinfo handle */
} rd_devstate_t;

typedef struct {
	int     bufsize;
	caddr_t bufaddr;
	}buf_info;

typedef struct {
	FIELD_INFO *field;
	caddr_t    bufaddr;
	}FLD_REC;
/*
 * An opaque handle where our set of ramdisk devices lives
 */
static void *rd_state;

static int rd_open(dev_t *devp, int flag, int otyp, cred_t *cred);
static int rd_close(dev_t dev, int flag, int otyp, cred_t *cred);
static rd_ioctl(dev_t dev, int cmd, int arg, int mode, cred_t * cred_p, int *rval_p);
static int rd_strategy(struct buf *bp);
static int rd_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int rd_write(dev_t dev, struct uio *uiop, cred_t *credp);
static int rd_mmap(dev_t dev, off_t off, int prot);
static int rd_print(dev_t dev, char *str);
static int rd_poll(dev_t dev, short events, int anyyet,
	short *reventsp, struct pollhead **phpp);

static struct cb_ops rd_cb_ops = {
	rd_open,
	rd_close,	/* close */
	rd_strategy,
	rd_print,
	nodev,		/* dump */
	rd_read,
	rd_write,
	rd_ioctl,	/* ioctl */
	nodev,		/* devmap */
	rd_mmap,	/* mmap */
	ddi_segmap,		/* segmap */
	rd_poll,	/* poll */
	ddi_prop_op,
	NULL,
	D_NEW | D_MP
};

static int rd_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
    void **result);
static int rd_identify(dev_info_t *dip);
static int rd_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int rd_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

static struct dev_ops rd_ops = {
	DEVO_REV,
	0,
	rd_getinfo,
	rd_identify,
	nulldev,	/* probe */
	rd_attach,
	rd_detach,
	nodev,		/* reset */
	&rd_cb_ops,
	(struct bus_ops *)0
};


extern struct mod_ops mod_driverops;

static struct modldrv modldrv = {
	&mod_driverops,
	"RTV-ramdisk driver v1.0",
	&rd_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	0
};

int
_init(void)
{
	int e;

	if ((e = ddi_soft_state_init(&rd_state,
	    sizeof (rd_devstate_t), 1)) != 0) {
		return (e);
	}

	if ((e = mod_install(&modlinkage)) != 0)  {
		ddi_soft_state_fini(&rd_state);
	}

	return (e);
}

int
_fini(void)
{
	int e;

	if ((e = mod_remove(&modlinkage)) != 0)  {
		return (e);
	}
	ddi_soft_state_fini(&rd_state);
	return (e);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

static int
rd_identify(dev_info_t *dip)
{
	if (strcmp(ddi_get_name(dip), "rtvramdisk") == 0) {
		return (DDI_IDENTIFIED);
	} else
		return (DDI_NOT_IDENTIFIED);
}

static int
rd_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int nblocks;
	int instance;
	register int i;
	char rawname[15], blkname[12];
	rd_devstate_t *rsp;
	RAM_FIELD_QUE *ram_shm_que;

	switch (cmd) {

	case DDI_ATTACH:

		instance = ddi_get_instance(dip);

		if (ddi_soft_state_zalloc(rd_state, instance) != DDI_SUCCESS) {
			cmn_err(CE_CONT, "%s%d: can't allocate state\n",
			    ddi_get_name(dip), instance);
			return (DDI_FAILURE);
		} else
			rsp = ddi_get_soft_state(rd_state, instance);


		/*
		 * Assign shared memory address and tables to softc (rsp).
		 */
		rsp->user_mapped=0;
		rsp->maxphys=ramdisk_maxphys;
		rsp->ramsize= MAX_RAM_SIZE;
		rsp->start_record= rsp->start_play=0;
		/*
		 * Assign ramdisk memory address to shared memory pointer.
		 */
		ram_shm_que = kmem_alloc(sizeof(RAM_FIELD_QUE), KM_SLEEP);
#ifdef UseSoftIntr
		ram_shm_que->base_addr_in=kmem_alloc(rsp->ramsize,KM_SLEEP);
#endif
		rsp->ram_shm_que=ram_shm_que;

		/*
		 * The 'nblocks' property is mandatory for block devices
		 */
		nblocks = rsp->ramsize / DEV_BSIZE;
		if (ddi_prop_create(DDI_DEV_T_NONE, dip, DDI_PROP_CANSLEEP,
		    "nblocks", (caddr_t)&nblocks, sizeof (int))
		    != DDI_PROP_SUCCESS) {
			cmn_err(CE_CONT, "%s%d: can't create nblocks prop\n",
			    ddi_get_name(dip), instance);
			goto attach_failed;
		}

		sprintf(rawname, "%d,raw", instance);
		sprintf(blkname, "%d", instance);
		if ((ddi_create_minor_node(dip,
		    rawname, S_IFCHR, instance, NULL, NULL) == DDI_FAILURE) ||
		    (ddi_create_minor_node(dip,
		    blkname, S_IFBLK, instance, NULL, NULL) == DDI_FAILURE)) {
			ddi_remove_minor_node(dip, NULL);
			goto attach_failed;
		}
#ifdef UseSoftIntr
	        if(ddi_add_softintr(dip, DDI_SOFTINT_HIGH, &rsp->put_ram_id,
		        &rsp->iblk_put_ram_cookie,(ddi_idevice_cookie_t *) 0, 
		        ramdisk_put_ram_intr, (caddr_t)rsp) != DDI_SUCCESS) 
		{
				return(DDI_FAILURE);
		}
	        if(ddi_add_softintr(dip, DDI_SOFTINT_HIGH, &rsp->get_ram_id,
		        &rsp->iblk_get_ram_cookie,(ddi_idevice_cookie_t *) 0, 
		        ramdisk_get_ram_intr, (caddr_t)rsp) != DDI_SUCCESS) 
		{
				return(DDI_FAILURE);
		}
		mutex_init(&rsp->put_ram_intr_mutex, "ramdisk put ram mutex",
					MUTEX_DRIVER, rsp->iblk_put_ram_cookie);
		mutex_init(&rsp->get_ram_intr_mutex, "ramdisk get ram mutex",
					MUTEX_DRIVER, rsp->iblk_get_ram_cookie);
#endif

		rsp->dip = dip;
		ddi_report_dev(dip);
		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}

attach_failed:
	/*
	 * Use our own detach routine to toss
	 * away any stuff we allocated above.
	 */
	(void) rd_detach(dip, DDI_DETACH);
	return (DDI_FAILURE);
}

static int
rd_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int instance;
	register rd_devstate_t *rsp;
	RAM_FIELD_QUE *ram_shm_que;

	switch (cmd) {

	case DDI_DETACH:
		/*
		 * Undo what we did in rd_attach, freeing resources
		 * and removing things we installed.  The system
		 * framework guarantees we are not active with this devinfo
		 * node in any other entry points at this time.
		 */
		instance = ddi_get_instance(dip);
		rsp = ddi_get_soft_state(rd_state, instance);
#ifdef UseSoftIntr
		mutex_destroy(&rsp->put_ram_intr_mutex);
		mutex_destroy(&rsp->get_ram_intr_mutex);
		ddi_remove_softintr(rsp->put_ram_id);
		ddi_remove_softintr(rsp->get_ram_id);
#endif

		ram_shm_que= rsp->ram_shm_que;
		if (ram_shm_que!=NULL && rsp->ramsize)
		{
#ifdef UseSoftIntr
			kmem_free(ram_shm_que->base_addr_in, rsp->ramsize);
#endif
		        kmem_free(ram_shm_que,sizeof(RAM_FIELD_QUE));
		}
		ddi_prop_remove_all(dip);
		ddi_remove_minor_node(dip, NULL);
		ddi_soft_state_free(rd_state, instance);
		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}
}

/*ARGSUSED*/
static int
rd_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	rd_devstate_t *rsp;
	int error = DDI_FAILURE;

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if ((rsp = ddi_get_soft_state(rd_state,
		    getminor((dev_t) arg))) != NULL) {
			*result = rsp->dip;
			error = DDI_SUCCESS;
		} else
			*result = NULL;
		break;

	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)getminor((dev_t) arg);
		error = DDI_SUCCESS;
		break;

	default:
		break;
	}

	return (error);
}

static int
rd_close(dev_t dev, int flag, int otyp, cred_t *cred)
{
	register rd_devstate_t *rsp;
	struct buf *bp;
        RAM_FIELD_QUE *ram_shm_que;
	int instance = getminor(dev);
	rsp = ddi_get_soft_state(rd_state, instance);
	rsp->rtv_open=0;
	rsp->start_play=0;
	rsp->start_record=0;
	if(rsp->user_mapped)
	{
		bp = &rsp->userbuf;
	        bp_mapout(bp);
		rsp->user_mapped=0;
	}
	return(0);
}
static int
rd_open(dev_t *devp, int flag, int otyp, cred_t *cred)
{
int err;
register rd_devstate_t *rsp;
int instance;
int status;
dev_t dsk;
SHM_FIELD_QUE *queue = global_shm_que;
RAM_FIELD_QUE *ram_shm_que;
QUEUE_INFO *ram_input_que,*ram_output_que;

	dev_info_t *sd_dip = (dev_info_t *)NULL;
	caddr_t addr;
	dev_t sd_dev;
	register int sd_instance;

	if (otyp != OTYP_BLK && otyp != OTYP_CHR)
		return (EINVAL);

	if (ddi_get_soft_state(rd_state, getminor(*devp)) == NULL)
		return (ENXIO);
	instance = getminor(*devp);
	rsp = ddi_get_soft_state(rd_state, instance);
	rsp->que_size_in=0;
	rsp->que_size_out=0;
	rsp->start_record= rsp->start_play=0;
	rsp->blocking_write=0;
	rsp->blocking_read=0;
	rsp->rtv_open=1;
	rsp->interlace=1;
        ram_shm_que=rsp->ram_shm_que;
	ram_input_que= (QUEUE_INFO *)&ram_shm_que->io_stream[0];
	ram_input_que->head= ram_input_que->que;
	ram_input_que->tail= &ram_input_que->que[MAX_FIELD_QUE_SIZE-1];
	ram_input_que->input=	ram_input_que->output=	ram_input_que->head;
	ram_input_que->level=0;
	ram_output_que=(QUEUE_INFO *)&ram_shm_que->io_stream[1];
	ram_output_que->head= ram_output_que->que;
	ram_output_que->tail= &ram_output_que->que[MAX_FIELD_QUE_SIZE-1];
	ram_output_que->input=	ram_output_que->output=	ram_output_que->head;
	ram_output_que->level=0;
	ram_shm_que->input_offset=0;
	ram_shm_que->output_offset=0;
#ifdef UseSoftIntr
	/*
	 * Copy rsp->put_ram_id to global variable;
	 */
	queue->ramdisk_get_trig= rsp->get_ram_id;
	queue->ramdisk_put_trig= rsp->put_ram_id;
#endif
	return (0);
}
static void
rd_minphys(struct buf *bp)
{
	rd_devstate_t *rsp;

	rsp = ddi_get_soft_state(rd_state, getminor(bp->b_edev));
	if (bp->b_bcount > rsp->maxphys)
		bp->b_bcount = rsp->maxphys;
}
static int send_audio_cmd(queue_t *wq)
{
mblk_t *mp;
struct iocblk *iocp;
int frame_no;

	    mp= allocb(4,BPRI_MED);
	    if(mp != NULL)
	    {
	            mp->b_datap->db_type=M_IOCTL;
	            iocp = (struct iocblk *)mp->b_rptr;
		    iocp->ioc_cmd=I_SET_AUDIO_RATE;
		    iocp->ioc_count=1;
		    putq(wq,mp);
	    }
}

static int que_frame(queue_t *wq)
{
mblk_t *mp;
struct iocblk *iocp;
int frame_no;

	    mp= allocb(4,BPRI_MED);
	    if(mp != NULL)
	    {
	            mp->b_datap->db_type=M_IOCTL;
	            iocp = (struct iocblk *)mp->b_rptr;
		    iocp->ioc_cmd=QUEUE_AUDIO_FRAME;
		    iocp->ioc_count=1;
		    putq(wq,mp);
	    }
}

typedef struct {
	FIELD_INFO *field;
	caddr_t addr;
	}USER_INFO;

typedef struct {
	int length;
	int sync_clock;
	caddr_t address;
	}AudioRec;
static int
rd_ioctl(dev_t dev, int cmd, int arg, int mode, cred_t * cred_p, int *rval_p)
{
register rd_devstate_t *rsp;
register int stat;
register int ret;
RAM_FIELD_QUE *ram_shm;
START_INFO *play_info;
QUEUE_INFO *ram_qinfo;
FIELD_INFO *field;
caddr_t addr;
int frame_size;
AUDIO_FRAME *audio_frame;
QUEUE_INFO *playback_qinfo;
AudioRec *audio_rec;
SHM_FIELD_QUE *queue = global_shm_que;
AUDIO_QUEUE_INFO *audio_que= (AUDIO_QUEUE_INFO *)queue->audio_shm;
int instance = getminor(dev);
rsp = ddi_get_soft_state(rd_state, instance);

	switch(cmd)
	{
	case START_RECORD:
	     rsp->start_record=1;
	     break;
	case STOP_RECORD:
	     rsp->start_record=0;
	     break;
	case START_PLAY:
	     rsp->start_play=1;
	     break;
	case STOP_PLAY:
	     rsp->start_play=0;
	     break;
	case RD_SET_INTERLACE:
	     rsp->interlace= 1;
	     break;
	case RD_SET_NONINTERLACE:
	     rsp->interlace= 0;
	     break;
	case RD_SET_AUDIO_RATE:
	     send_audio_cmd((queue_t *)global_shm_que->aud_wq);
	     break;
	case RD_PUT_AUDIO:
	     {
		caddr_t inptr;

		ASSERT(arg);
		audio_rec= (AudioRec *)arg;


		if(audio_que->mem_offset+audio_rec->length>= AUDIO_RAM_SIZE)
		    audio_que->mem_offset=0;

		inptr= audio_rec->address;
		audio_frame = audio_que->input;
		audio_frame->size= audio_rec->length;
		queue->audio_sync = audio_rec->sync_clock;
		audio_frame->addr = 
		    (caddr_t)(audio_que->base+audio_que->mem_offset);
		audio_que->mem_offset+= audio_rec->length;
		if(ddi_copyin(inptr,audio_frame->addr,audio_rec->length,0) == -1)
		{
		        cmn_err(CE_WARN,"! put_audio: ddi_copyin error\n");
		        return(EIO);
		}
		audio_que->input++;
		audio_que->level++;
		if(audio_que->input>= audio_que->tail)
			audio_que->input= audio_que->head;
		que_frame((queue_t *)global_shm_que->aud_wq);
		return(0);
	     }
	     break;
	case RD_GET_AUDIO:
	     {
		 if(audio_que->level<2)
		         return(EIO);

	         if(audio_que->output!=audio_que->input)
	         {
		     caddr_t outptr;
		     register int bufsize;
		     audio_rec = (AudioRec *)arg;

		     outptr= audio_rec->address;
		     audio_frame= audio_que->output;
		     bufsize= audio_rec->length;
		     audio_rec->sync_clock= audio_frame->sync_clock;
	             if(ddi_copyout(audio_frame->addr,outptr,bufsize,0)== -1)
	             {
	                    cmn_err(CE_WARN,"!get_audio: ddi_copyout failed\n");
	                    return(EIO);
	             }
		     audio_que->output+= 1;
		     audio_que->level-= 1;
		     if(audio_que->output>= audio_que->tail)
		             audio_que->output=audio_que->head;
	        }
		return(0);
	     }
	     break;
	case RD_GET_RECORD_LEVEL:
	     ddi_copyout((caddr_t)&queue->record_qlevel,(caddr_t)arg,4,0);
	     return(0);
	     break;
	case RD_GET_PLAY_LEVEL:
#ifdef UseSoftIntr
	     ram_shm= rsp->ram_shm_que;
	     ram_qinfo= (QUEUE_INFO *)&ram_shm->io_stream[1];
	     ddi_copyout((caddr_t)&ram_qinfo->level,(caddr_t)arg,4,0);
#endif
	     playback_qinfo = &queue->playback_streams[rsp->stream_id];
             ddi_copyout((caddr_t)&playback_qinfo->level,(caddr_t)arg,4,0);
	     return(0);
	     break;
	case RD_GET_FIELD:
#ifdef UseSoftIntr
		ret= get_ram_field(rsp,arg);	
#endif
		ret= copy_shm_to_user(rsp,arg);
		return(ret);
	     break;
	case RD_PUT_FIELD:
#ifdef UseSoftIntr
		ret=put_ram_field(rsp,arg);
#endif
		ret= copy_usr_to_shm(rsp,arg);
		/*
	        drv_usecwait(1000);
		*/
		return(ret);
	     break;
        case RD_AUDIO_MONITOR_ON:
	     global_shm_que->audio_monitor=1;
	     break;
	case RD_AUDIO_MONITOR_OFF:
	     global_shm_que->audio_monitor=0;
	     break;
	case RD_START_AUDIO_RECORD:
	     global_shm_que->audio_mode= RECORD;
	     break;
	case RD_START_AUDIO_PLAYBACK:
	     global_shm_que->audio_mode= PLAY;
	     break;
	case RD_STOP_AUDIO_RECORD:
	case RD_STOP_AUDIO_PLAYBACK:
	     global_shm_que->audio_mode=0;
	     break;
	case INIT_QUEUE:
	     audio_que->input=audio_que->output=audio_que->head;
	     audio_que->level=0;
	     break;
	case SET_OUTPUT_STREAM:
	     play_info= (START_INFO *)arg;
	     rsp->stream_id= play_info->wid;
	     break;
	}
	
	return(0);
}

/*ARGSUSED*/
static int
rd_read(dev_t dev, struct uio *uiop, cred_t *credp)
{
	int err;
	int retval;
	struct buf *bp;
	int instance = getminor(dev);
	rd_devstate_t *rsp = ddi_get_soft_state(rd_state, instance);

	bp= &rsp->userbuf;
	retval=physio(rd_strategy,bp,dev,B_READ,rd_minphys,uiop);

	return(retval);
}

/*ARGSUSED*/
static int
rd_write(dev_t dev, register struct uio *uiop, cred_t *credp)
{
	int retval;
	struct buf *bp;
	int instance = getminor(dev);
	rd_devstate_t *rsp = ddi_get_soft_state(rd_state, instance);

	bp= &rsp->userbuf;
	retval = physio(rd_strategy, bp, dev, B_WRITE, rd_minphys, uiop);

	return(retval);
}

static int
rd_strategy(struct buf *bp)
{
	register int field_size,status;
	register caddr_t raddr,buf_addr;
	rd_devstate_t *rsp;

	rsp = ddi_get_soft_state(rd_state, getminor(bp->b_edev));
	if (rsp == NULL) {
		bp->b_error = ENXIO;
		bp->b_flags |= B_ERROR;
	} else {
		register int retval;
		register caddr_t	buf_addr, raddr ;

		if (bp->b_flags & B_PAGEIO)
		{
			bp_mapin(bp);
		        rsp->user_mapped=1;
		}
	}
	biodone(bp);
	return (0);
}
rd_print(dev_t dev, char *str)
{
	int instance = getminor(dev);
	rd_devstate_t *rsp = ddi_get_soft_state(rd_state, instance);

	cmn_err(CE_WARN, "%s%d: %s\n", ddi_get_name(rsp->dip), instance, str);
	return (0);
}

rd_poll(dev_t dev, short events, int anyyet, short *reventsp, struct pollhead **phpp)
{
	printf("rd_poll\n");
}
rd_mmap(dev_t dev,off_t off, int prot)
{
RAM_FIELD_QUE *ram_shm_que;
int instance = getminor(dev);
rd_devstate_t *rsp = ddi_get_soft_state(rd_state, instance);

	ram_shm_que= rsp->ram_shm_que;

#ifdef NotInUse
	return( hat_getkpfnum((caddr_t)ram_shm_que->base_addr_in));
#endif
}
/*
#define TRACE_GETRAM_LEVEL
#define TRACE_PUT_INTR
*/

static int
copy_shm_to_user(rd_devstate_t *rsp,caddr_t buf)
{
FIELD_INFO *shmfld,*field;
register int rec_size,bytcnt;
caddr_t field_address;
caddr_t bufaddr;
static int queue_empty= -1;
static int trace=0;
SHM_FIELD_QUE *queue = global_shm_que;
	/* 
	 * Copy out the field information.
	 */
	 shmfld= queue->record_output;
	 bufaddr= buf;
	 if(ddi_copyout((caddr_t)shmfld,bufaddr,sizeof(FIELD_INFO),0)== -1)
	 {
	        cmn_err(CE_WARN,"copy_shm: copyout field info failed\n");
	        return(EIO);
	 }
	/* 
	 *  Get the shared memory field data address.
	 *  Copy the field data.
	 */
	 bufaddr += sizeof(FIELD_INFO);
	 field_address = (caddr_t)(queue->dma_in_addr+shmfld->offset);
	 if(ddi_copyout(field_address,bufaddr,shmfld->size,0) == -1)
	 {
		cmn_err(CE_WARN,"copy_shm: copyout data sz=%d error!\n",
			shmfld->size);
		return(EIO);
	 }
	/*
	 *  Adjust the queue pointer and input level.
	 */
	 shmfld->inuse=0;
	 queue->record_output++;
	 queue->record_qlevel--;
	 if (queue->record_output >= queue->record_tail)
	     queue->record_output = queue->record_head;
	 return(0);
}
static int
copy_usr_to_shm(rd_devstate_t *rsp, caddr_t buf)
{
QUEUE_INFO *playback_qinfo;
FIELD_INFO *field;
FIELD_INFO *put_field;
caddr_t field_address;
caddr_t bufaddr;
register int bytcnt,rec_size;
SHM_FIELD_QUE *queue = global_shm_que;
static int trace=0;

	if (!queue) return(EIO);  /* change to assert in production */
	/*
	 * Get local pointer to playback queue.
	 */
	playback_qinfo = &queue->playback_streams[rsp->stream_id];
	if(playback_qinfo->level>30)
	{
		cmn_err(CE_WARN,"!Queue %d Full\n",rsp->stream_id);
		return(EIO);
	}
	/*
	if(playback_qinfo->input!=playback_qinfo->output)
	*/
	{
	    /* 
	     * Get pointer to frame playback queue
	     */
	    field = (FIELD_INFO *) playback_qinfo->input;
	    /*
	     * Test for dma_out_offset overflow 
	     * and reset to base_offset if necessary.
	     */
	    bufaddr= buf;
	    if(ddi_copyin((caddr_t)bufaddr,(caddr_t)field,sizeof(FIELD_INFO),0) == -1)
	    {
		    cmn_err(CE_WARN,"copy_usr: copyin field info\n");
		    return(EIO);
	    }
	    if(queue->dma_out_offset[0] + field->size >= queue->max_shm_size)
		{
	        queue->dma_out_offset[0]=0;
		}

	    bufaddr += sizeof(FIELD_INFO);
	    field->offset = queue->dma_out_offset[0];
	    field_address = (caddr_t)(queue->dma_out_addr+field->offset);

	    if(ddi_copyin(bufaddr,field_address,field->size,0)== -1)
	    {
		    cmn_err(CE_WARN,"copy_usr: copyin field 0 data!\n");
		    return(EIO);
	    }
	    queue->dma_out_offset[0] += field->size;
	    /* 
	     * Advance playback queue level and input pointers, 
	     * reset if necessary
	     */
	    playback_qinfo->input++;
	    playback_qinfo->level+=1;
	    if (playback_qinfo->input >= playback_qinfo->tail)
		    playback_qinfo->input = playback_qinfo->head;
	}

	return(0);
}
#ifdef UseSoftIntr
static int 
get_ram_field(rd_devstate_t *rsp, caddr_t buf)
{
	RAM_FIELD_QUE *ram_shm;
	QUEUE_INFO *ram_qinfo;
	FIELD_INFO *field;
	caddr_t addr;
	register int rec_size;
	static int trace=0;

	ram_shm=rsp->ram_shm_que;
	ram_qinfo= (QUEUE_INFO *)&ram_shm->io_stream[0];
	if(ram_qinfo->level<=0)
	{
#ifdef TRACE_QUEUE_EMPTY
	    cmn_err(CE_WARN,"!Queue Empty\n");
#endif
	    return(EIO);
	}
	if(ram_qinfo->input!=ram_qinfo->output)
	{
	    field= ram_qinfo->output;
	    if(!field->inuse)
	    {
	        ram_qinfo->output++;
	        ram_qinfo->level--;
	        if (ram_qinfo->output >= ram_qinfo->tail)
	            ram_qinfo->output = ram_qinfo->head;
		return(EAGAIN);
	    }
	    if(field->size[0]<=0 || field->size[1] <= 0)
	    {
		 cmn_err(CE_WARN,"!Invalid field->size[0]=%d\n",field->size[0]);
	         ram_qinfo->output++;
	         ram_qinfo->level--;
	         if (ram_qinfo->output >= ram_qinfo->tail)
	            ram_qinfo->output = ram_qinfo->head;
		 return(EINVAL);
	    }
	    if(ddi_copyout((caddr_t)field,buf,sizeof(FIELD_INFO),0)== -1)
	    {
	        cmn_err(CE_WARN,"get_ram: copyout field failed\n");
	        return(EIO);
	    }
#ifndef SHARED_MEMORY_SUPPORT
	    addr= (caddr_t)(ram_shm->base_addr_in+field->offset[0]);
	    rec_size= field->size[0]+field->size[1];
	    if( ddi_copyout(addr,(buf+sizeof(FIELD_INFO)),rec_size,0) == -1)
	    {
		cmn_err(CE_WARN,"get_ram_field: copyout data sz=%d error!\n",
			field->size);
		return(EIO);
	    }
#endif
	    ram_qinfo->output++;
	    ram_qinfo->level--;
	    if (ram_qinfo->output >= ram_qinfo->tail)
	        ram_qinfo->output = ram_qinfo->head;
#ifdef TRACE_GETRAM_LEVEL
	    if((field->frame_no%30)== 0)
	    {
	        cmn_err(CE_WARN,"!get_ram_field: frame= %d level=%d\n",
			field->frame_no,ram_qinfo->level);
	    }
#endif
	    return(0);
	}
	else
	{
		cmn_err(CE_WARN,"!Input == Output\n");
	}
	return(EIO);
}
static int
put_ram_field(rd_devstate_t *rsp, caddr_t buf)
{
	RAM_FIELD_QUE *ram_shm;
	QUEUE_INFO *ram_qinfo;
	FIELD_INFO *field;
	caddr_t addr;
	static int trace=0;
	register int rec_size;

	ram_shm= rsp->ram_shm_que;
	ram_qinfo= (QUEUE_INFO *)&ram_shm->io_stream[1];
	if(ram_qinfo->level>=MAX_FIELD_QUE_SIZE)
	{
	     cmn_err(CE_WARN,"put_ram_field: qlevel is %d\n",
		ram_qinfo->level);
	     return(EAGAIN);
	}
	field= ram_qinfo->input;
	if(ddi_copyin(buf,(caddr_t)field,sizeof(FIELD_INFO),0)== -1)
	{
		cmn_err(CE_WARN,"put_ram_field: copyin field error!\n");
		return(EIO);
	}

#ifndef SHARED_MEMORY_SUPPORT
	rec_size= field->size[0]+ field->size[1];
	if(ram_shm->output_offset+rec_size >= MAX_RAM_SIZE)
	    ram_shm->output_offset=0;
	field->offset[0] =ram_shm->output_offset;
	addr= (caddr_t)(ram_shm->base_addr_in+field->offset[0]);
	if(ddi_copyin((buf+sizeof(FIELD_INFO)),addr,rec_size,0) == -1)
	{
		cmn_err(CE_WARN,"put_ram_field: copyin data sz=%d error!\n",
			field->size);
		return(EIO);
	}
	ram_shm->output_offset+= rec_size;
#endif
	ram_qinfo->input++;
	ram_qinfo->level++;
	if(ram_qinfo->input>= ram_qinfo->tail)
	        ram_qinfo->input= ram_qinfo->head;
	
	return(0);
}
static int
copy_ram_to_shm(rd_devstate_t *rsp, FIELD_INFO *ramfld,caddr_t bufaddr)
{
QUEUE_INFO *playback_qinfo;
FIELD_INFO *field;
FIELD_INFO *put_field;
caddr_t field_address,ram_addr;
register int bytcnt,rec_size;
SHM_FIELD_QUE *queue = global_shm_que;
static int trace=0;

	if (!queue) return(-1);  /* change to assert in production */
	/*
	 * get local pointer to playback queue.
	 */
	playback_qinfo = &queue->playback_streams[rsp->stream_id];
	/* 
	 * get pointer to frame playback queue
	 */
	field = (FIELD_INFO *) playback_qinfo->input;
	/* 
	 * copy frame parameters from ramfld arg.
	 */
	bcopy((caddr_t)ramfld,(caddr_t)field,sizeof(FIELD_INFO));
	rec_size= field->size[0] + field->size[1];
	/*
	 * Test for out_offset overflow and reset to base_offset if necessary.
	 */
	if(queue->dma_out_offset[0]+rec_size >= queue->max_shm_size)
	     queue->dma_out_offset[0]=0;
	/* 
	 * Get field offset from shared memory dma address 
	 */
	field->offset[0] = queue->dma_out_offset[0];
	/*
	 * Set shared memory address & Copy field 0 data.
	 */
	field_address = (caddr_t)(queue->dma_out_addr+field->offset[0]);
	bytcnt= field->size[0];
	bcopy(bufaddr,field_address,bytcnt);
	if(rsp->interlace)
	{
	/*
	 * Set field 1 offset copy field 1 data.
	 */
	    field->offset[1] = field->offset[0] + bytcnt;
	    ram_addr= (bufaddr+field->size[0]);
	    bcopy(ram_addr,(caddr_t)(field_address+bytcnt),field->size[1]);
	}
	/* 
	 * increment dma output offset by current record size.
	 */
	 queue->dma_out_offset[0] += rec_size;
	/* 
	 * Advance playback queue level and input pointers, 
	 * reset if necessary
	 */
	playback_qinfo->input++;
	playback_qinfo->level++;
	if (playback_qinfo->input >= playback_qinfo->tail)
		playback_qinfo->input = playback_qinfo->head;

	return(1);
}

static u_int
ramdisk_get_ram_intr(rd_devstate_t *rsp)
{
register int offset;
SHM_FIELD_QUE *queue = global_shm_que;
RAM_FIELD_QUE *ram_shm;
QUEUE_INFO *ram_qinfo,*playback_qinfo;
FIELD_INFO *ramfld,*shmfld;
caddr_t baddr;
static int trace=0;
static int trace1=0;

	mutex_enter(&rsp->get_ram_intr_mutex);
#ifdef  GET_RAM_TRIGGER
	if(rsp->start_play == 1 && queue->get_trigger==GETRAM)
	{
	    queue->get_trigger=0;
	    ram_shm= rsp->ram_shm_que;
	    ram_qinfo= (QUEUE_INFO *)&ram_shm->io_stream[1];
	    if(ram_qinfo->level<=0 )
	    {
		mutex_exit(&rsp->get_ram_intr_mutex);
		return (DDI_INTR_CLAIMED);
	    }
	    if(ram_qinfo->input!=ram_qinfo->output)
	    {
	        ramfld= ram_qinfo->output;
		/*
		 * Use input address as share I/O buffer for 
		 * testing. Later this should be changed 
		 * to use "base_addr_out" (second shared memory buffer).
		 */
	        baddr= (caddr_t)(ram_shm->base_addr_in+ramfld->offset[0]);
	        if(copy_ram_to_shm(rsp,ramfld,baddr)== 1)
		{
	            ram_qinfo->level--;
	            ram_qinfo->output++;
		    if(ram_qinfo->output>=ram_qinfo->tail)
		        ram_qinfo->output=ram_qinfo->head;
		}
	    }
	}
#endif
	mutex_exit(&rsp->get_ram_intr_mutex);
	return (DDI_INTR_CLAIMED);
}

static int
copy_shm_to_ram(rd_devstate_t *rsp,FIELD_INFO *ramfld)
{
FIELD_INFO *shmfld;
RAM_FIELD_QUE *ram_shm;
register int rec_size,bytcnt;
caddr_t field_address;
caddr_t ram_address;
SHM_FIELD_QUE *queue = global_shm_que;
	/*
	 * Assign current shared memory queue frame address,
	 * and the ram shared memory address.
	 */
	 shmfld = queue->record_output;
	 if(!shmfld->inuse)
	 {
	     queue->record_output++;
	     if (queue->record_output >= queue->record_tail)
	         queue->record_output = queue->record_head;
	     return 0;
	 }
	 ram_shm= rsp->ram_shm_que;
	/* 
	 * Copy the frame descriptor.
	 */
	 bcopy((caddr_t)shmfld,(caddr_t)ramfld,sizeof(FIELD_INFO));
	 rec_size= ramfld->size[0]+ ramfld->size[1];
	 /*
	  * Test for ram buffer overflow and reset offset
	  * if necessary.
	  */
	 if(ram_shm->input_offset+rec_size>=rsp->ramsize)
	    ram_shm->input_offset=0;
	 ramfld->offset[0] = ram_shm->input_offset;
	 ram_address= ram_shm->base_addr_in+ramfld->offset[0];
	 ram_shm->input_offset+= rec_size;
	/* 
	 *  Get the ram and shared memory field 0 data address.
	 *  Copy the field 0 data.
	 */
	 field_address = (caddr_t)(queue->dma_in_addr+shmfld->offset[0]);
	 bytcnt= shmfld->size[0];
	 bcopy(field_address,ram_address,bytcnt);
	 if(rsp->interlace)
	 {
	/* 
	 *  Get the field 1 data address.
	 *  Copy the field 1 data.
	 */
	     field_address = (caddr_t)(queue->dma_in_addr+shmfld->offset[1]);
	     bcopy(field_address,(caddr_t)(ram_address+bytcnt),shmfld->size[1]);
	     ramfld->offset[1]= ramfld->offset[0]+ramfld->size[0];
	 }
	/*
	 *  Adjust the queue pointer and input level.
	 */
	 shmfld->inuse=0;
	 queue->record_output++;
	 queue->record_qlevel--;
	 if (queue->record_output >= queue->record_tail)
	     queue->record_output = queue->record_head;
	 return(rec_size);
}
static u_int
ramdisk_put_ram_intr( rd_devstate_t *rsp)
{
SHM_FIELD_QUE *queue = global_shm_que;
RAM_FIELD_QUE *ram_shm;
QUEUE_INFO *ram_qinfo;
FIELD_INFO *ramfld,*shmfld;
register int rec_size;
caddr_t baddr;
static int trace=0;
static int trace1=0;

	mutex_enter(&rsp->put_ram_intr_mutex);

#ifdef  PUT_RAM_TRIGGER
	if(rsp->start_record == 1 && queue->put_trigger==PUTRAM)
	{
	    queue->put_trigger=0;
	    ram_shm= rsp->ram_shm_que;
	    if(queue->record_qlevel<=0)
	    {
		mutex_exit(&rsp->put_ram_intr_mutex);
		return (DDI_INTR_CLAIMED);
	    }
	    ram_qinfo= (QUEUE_INFO *)&ram_shm->io_stream[0];
	    if(queue->record_input != queue->record_output)
	    {
		if(ram_qinfo->level>=MAX_FIELD_QUE_SIZE) 
		{
		    mutex_exit(&rsp->put_ram_intr_mutex);
		    return (DDI_INTR_CLAIMED);
		}
	        ramfld= ram_qinfo->input;
	        if((rec_size= copy_shm_to_ram(rsp,ramfld))==0)
	        {
		    mutex_exit(&rsp->put_ram_intr_mutex);
		    return (DDI_INTR_CLAIMED);
	        }
	        ram_qinfo->input++;
	        ram_qinfo->level++;
#ifdef TRACE_PUT_INTR
	if(trace++ == 30)
	{
	cmn_err(CE_WARN,"!put_ram_intr: frame=%d level=%d\n",
			ramfld->frame_no,ram_qinfo->level);
			trace=0;
	}
#endif
		if(ram_qinfo->input>=ram_qinfo->tail)
		    ram_qinfo->input=ram_qinfo->head;
	    }
	}
#endif
	mutex_exit(&rsp->put_ram_intr_mutex);
	return (DDI_INTR_CLAIMED);
}
#endif


