/***********************************************************
Copyright 1991 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/*
** File: 
**
**   xvplx.c --- Xv Parallax device dependent module.
**
** Author: 
**
**   David Carver (Digital Workstation Engineering/Project Athena)
**
** Revisions:
**
**   26.02.92 Dave Arnold
**     - Initiall Parallax port
**
**   26.11.91 Carver
**     - optimized PutStill to not redraw enable plane between like stills.
**     - fixed multi-screen support
**     - added new call to pip_installed()
**
**   18.09.91 Carver
**     - changed interface to libpip.c; now uses pip_src_area and pip_dst_area
**       instead of pip_source_area, pip_origin, and pip_size.
**     - fixed constrain procedure to clip video destination to screen bounds
**
**   10.09.91 Carver
**     - fixed calls to fill_video_enable some were overrunning bounds
**       of screen.
**
**   29.08.91 Carver
**     - change: unrealizing windows no longer preempts video
**     - added support for duty cycle (video priority) management
**     - added support for video in StaticGray windows
**     - added include of cfb.h; needed to access 8 plane gc priv structure
**     - changed interfaces to rop.c and libpip.c to use RopPtr uniformly
**     - changed ultrix conditional compile so this file works for both
**       sun and dec boards
**
**   26.06.91 Carver
**     - fixed GC wrapping logic in a big way
**     - fixed XvropValidateGC; it had the wrong argument definition
**       can't believe that it ever worked.
**     - fixed XvropStopVideo to check to see if the GC still exists
**       before it references a pointer to it.
**
**   11.06.91 Carver
**     - changed SetPortControl to SetPortAttribute
**     - changed GetPortControl to GetPortAttribute
**     - changed QueryBestSize
**
**   04.06.91 Carver
**     - completed implementing port controls
**     - used code from R. Ulichney to to contrast and brightness calc's
**     - changed interface to libpip.c, use new libpip.h header file
**
**   31.05.91 Carver
**     - made big fixes to occlusion stuff
**     - fixed video image offset and size
**
**   23.05.91 Carver
**     - check for pip existance before initializing the adaptor
**
**   15.05.91 Carver
**     - version 2.0 upgrade
**
**   24.04.91 Carver
**     - updated to use 24 plane support.
**
**   19.03.91 Carver
**     - origional version coded to v1r4.
**
*/

#include <stdio.h>
#include <sys/types.h>

#define PSZ	32
#include "X.h"
#include "Xproto.h"
#include "scrnintstr.h"
#include "validate.h"
#include "windowstr.h"
#include "pixmapstr.h"
#include "gcstruct.h"
#include "dixstruct.h"
#include "cfb.h"                


#include "Xv.h"
#include "Xvproto.h"
#include "xvdix.h"
#include "tv2reg.h"
#include "plxInit.h"

#define M8VOFF 0x40000
#define M8SIZE (1152*900)
#define M24VOFF 0x140000
#define M24SIZE (1152*900*4)

#define GLOBAL
#include "xvplx.h"
#undef GLOBAL

#include "plxv.h"
#include "plxv_xpoint.h"
#include "xvideo.h"

#define NTSC_BLANK      45
#define PAL_BLANK       50
#define InvertedStringID	0
#define NonInvertedStringID	1
#define BWStringID		2
#define ColorStringID		3
#define GenlockStringID		4
#define FreerunStringID		5
#define XV_WINDOW_TYPE		2

typedef struct _PLXPort {
    int         io_type;
    XV_knob_rec knob;
    } PLXPortRec;

typedef struct _VidValueArgs {
	INT16 select B16;
	INT16 sat B16;
	INT16 cont B16;
	INT16 hue B16;
	INT16 brit B16;
	INT16 pad0 B16;
	INT32 pad1 B32;
} xVidValueArgs;
#define sz_xVidValueArgs 16

typedef struct _VideoSelectArgs {
	INT16 channel B16;
	INT16 standard B16;
	INT16 format B16;
	INT16 type B16;
} xVideoSelectArgs;
#define sz_xVideoSelectArgs 8;

typedef struct _VideoArgs {
	INT16 vx B16, vy B16, vw B16, vh B16;	/* input video area */
	INT16 x B16, y B16, w B16, h B16;	/* windowing cord's */
} xVideoArgs;
#define sz_xVideoArgs 16

extern PLXPortRec plxvideoport[4];

extern int input1_port;
extern int output0_port;
int plxrestore;
int i;


/* INTERNAL */
#define	X_VideoLive		12
#define X_VideoSqueezeLive	41
#define X_VideoSqueezeStill	42

#define X_VideoFlipLive		51
#define X_VideoFlipStill	52
#define X_VideoFlipScale	53
#define X_VideoFlipSqueezeLive	54
#define X_VideoFlipSqueezeStill	55
#define X_VideoScaleOutput	101
#define X_VideoOutputSelect	104

static Bool XvplxCloseScreen();
static int XvplxQueryAdaptors();
static int XvplxInitAdaptor();
static int XvplxAllocatePort();
static int XvplxFreePort();
static int XvplxPutVideo();
static int XvplxPutStill();
static int XvplxGetVideo();
static int XvplxGetStill();
static int XvplxStopVideo();
static int XvplxSetPortAttribute();
static int XvplxGetPortAttribute();
static int XvplxQueryBestSize();


static void XvplxClipNotify();
static void XvplxCopyWindow();
static Bool XvplxUnrealizeWindow();
void XvplxWindowExposures();

static void XvplxDestroyGC();
static void XvplxValidateGC();

static XvPortPtr XvplxPorts[MAXSCREENS][2];
XV_struct *xv_ctx;
extern XV_struct *PlxGetContext();
TV2Priv *init_priv_tv2();
static int tv2_priv_cnt=0;


TV2Priv *SetTv2Priv(DrawablePtr pDraw, GCPtr pGC)
{
    plxGCPtr	  privGC = GetPlxGCPrivate(pGC);
/*    plxPrivWinPtr plxWinPtr = plxGetPwinPriv(pDraw); */
    TV2Priv* 	  tv2_priv;
    int  	  wid;

  if(privGC->video_priv==NULL)
  {
      tv2_priv = init_priv_tv2(pGC);
#ifdef NOTDEF
      tv2_priv->video_tag  = plxGetWid(pGC->pScreen,pDraw);
      tv2_priv->window_type=PLX_VIDEO;
#endif
      privGC->video_priv= (int *)tv2_priv;
  }
  else 
      tv2_priv= (TV2Priv *)privGC->video_priv;
  /* plxWinPtr->Video_Priv= (int *)tv2_priv; */
  return(tv2_priv);
}

TV2Priv *GetTv2Priv(GCPtr pGC)
{
int  i=0;
TV2Priv *tv2_priv;
plxGCPtr	privGC = GetPlxGCPrivate(pGC);

  if(privGC->video_priv)
	return((TV2Priv *)privGC->video_priv);
  return(NULL);
}

TV2Priv *XvplxSetupGC(GCPtr pGC,DrawablePtr pDraw,XvPortPtr pPort)
{
  XvEncodingPtr pe;
  XvplxPortPtr ppp;
  int port, mod;
  TV2Priv *tv2_priv;
  plxGCPtr	privGC = GetPlxGCPrivate(pGC);

  ppp = (XvplxPortPtr)pPort->devPriv.ptr;
  pe = ppp->pEncoding;
  mod = 0;

  tv2_priv = SetTv2Priv(pDraw,pGC);

  if(pDraw->type == DRAWABLE_WINDOW && tv2_priv->video_tag==0){
	tv2_priv->video_tag = plxGetWid(pGC->pScreen,pDraw);
        tv2_priv->window_type=PLX_VIDEO;
         privGC->video_priv= (int *)tv2_priv;
  }
  if(tv2_priv->channel != ppp->channel)
  {
    tv2_priv->channel = ppp->channel;
    mod = 1;
  }


  if(tv2_priv->standard != (pe->id >> 16))
  {
    tv2_priv->standard = pe->id >> 16;
    mod = 1;
  }

  if(tv2_priv->format != (pe->id & 0xffff))
  {
    tv2_priv->format = pe->id & 0xffff;
    mod = 1;
  }

  if(mod)
  {
   if(tv2_priv->channel ==  PLX_INPUT_0)
    port = PORT_D;
   else if (tv2_priv->channel == PLX_INPUT_1)
    port = input1_port;
   else
    port = output0_port;

   if(port == output0_port)
   {
    if(tv2_priv->standard == PLX_NTSC)
     plxv_output_config(port, "525");
    else
     plxv_output_config(port, "625");
   }
   else
   {
    if(tv2_priv->standard == PLX_NTSC)
     plxv_input_config(port, "525");
    else
     plxv_input_config(port, "625");
   }
   switch (tv2_priv->format)
   {
        case PLX_COMP:
            plxv_input_config(port, "COMPOSITE");
            break;
        case PLX_YC:
            plxv_input_config(port, "YC");
            break;
        case PLX_YUV:
            plxv_input_config(port, "YUV");
            break;
        case PLX_RGB:
            plxv_input_config(port, "RGB");
            break;
   }
  }
  return(tv2_priv);
}

#define USE_XV
#ifdef USE_XV

void XvDdxInit(pScreenInfo)
ScreenInfo *pScreenInfo;
{
ScreenPtr 	pScreen;
plxScreenPtr	pPlxPrivate;
int i_screen=0;
  

    for ( i_screen = 0; i_screen < pScreenInfo->numScreens; i_screen++ )
    {
	    pScreen = pScreenInfo->screens[i_screen];
	    pPlxPrivate =  GetPlxPrivate(pScreen);
	    if(pPlxPrivate != NULL) 
 		{
		    xv_ctx = PlxGetContext(pScreen);
  			XvplxScreenInit(pScreen);
  			plxVideoInit(pScreen, xv_ctx);
		}
	}

}
#endif /* USE_XV */

TV2Priv *
init_priv_tv2(pGC)
GCPtr pGC;
{
TV2Priv *priv;

  priv = (TV2Priv *)xalloc(sizeof(TV2Priv));
  priv->window_type = PLX_GRAPHICS;
  priv->channel = 0;
  priv->standard = PLX_NTSC;
  priv->format = PLX_COMP;
  priv->type = PLX_RGB24;
  priv->gov_tag=0;
  priv->gov_mode=0;
  priv->xdirection = 0;
  priv->ydirection = 0;
  priv->video_tag = 0;
  priv->qtab = NULL;
  priv->htab = NULL;
  priv->pGC = pGC;
  priv->pScreen =  pGC->pScreen;
  priv->pWid = 0;
  return(priv);
}

#ifdef USE_XV
int
XvplxScreenInit(pScreen)
ScreenPtr pScreen;

{
  XvplxScreenPtr props;
  XvScreenPtr pxvs;
  int status,fd;


  /* ALL XV DD SCREEN INIT PROCS NEED TO CALL THE DI SCREEN INIT PROC FIRST */

  XvScreenInit(pScreen);

  pxvs = (XvScreenPtr)pScreen->devPrivates[XvScreenIndex].ptr;

  if ((status = XvplxInitAdaptor(pScreen)) != Success)
    return status;

  /* XXX --- THESE SHOULD BE CHAINED FOR MULTIPLE ADAPTOR SUPPORT */

  pxvs->ddCloseScreen = XvplxCloseScreen;
  pxvs->ddQueryAdaptors = XvplxQueryAdaptors;

  if (!(props = (XvplxScreenPtr)xalloc(sizeof(XvplxPortRec))))
    {
      return BadAlloc;
    }
  pxvs->devPriv.ptr = (pointer)props;
#ifdef NO_PLX_VIDEO
  props->WindowExposures = pScreen->WindowExposures;
  pScreen->WindowExposures = XvplxWindowExposures;
#endif
  return Success;
}

static int
XvplxInitAdaptor(pScreen)

ScreenPtr pScreen;

{
  int id,status,count,np,ne,nd,nv,nf,tf,na,nvis;
  XvScreenPtr pxvs;
  XvAdaptorPtr pa,pas;
  XvPortPtr pp,pps;
  XvplxPortPtr ppp;
  XvEncodingPtr pe, pes, opes;
  XvFormatPtr pf,pfs;
  unsigned long *pv;
  VisualPtr pvis;
  DepthPtr pd;
  int nports;
  XV_struct *xv=xv_ctx;


  pxvs = (XvScreenPtr) pScreen->devPrivates[XvScreenIndex].ptr;


  if(xv->opt.board_code)
	++xvplx_num_adaptors;

  pas = (XvAdaptorPtr)xalloc(sizeof(XvAdaptorRec) * xvplx_num_adaptors);
  if (!pas)
    {
      FatalError("XvplxInitAdaptors: coudn't alloc Adaptor struct(s)\n");
    }
  pa = pas;


/* find out how may ports and encodings*/

  for (na = 0; na<xvplx_num_adaptors; na++)
    {
      if(na == 0)
        pes = (XvEncodingPtr)xalloc(sizeof(XvEncodingRec) * 3);
      else
	pes = (XvEncodingPtr)xalloc(sizeof(XvEncodingRec) * 6);

      if (!pes)
	{
	  FatalError("XvplxInitAdaptors: coudn't alloc Encoding struct\n");
	}

      pe = pes;

      for (ne = 0; ne < XvplxNumEncodings[na]; ne++)
	{
	  if(na == 0)
	    *pe = XvplxBaseEncodings[na][ne];
	  else
	    *pe = (xv->opt.board_code == OPT_1) ? XvplxOpt1Encodings[0][ne] :
						XvplxOpt2Encodings[0][ne];
	  pe->pScreen = pScreen;

	  pe++;
	}

      pps = (XvPortPtr)xalloc(sizeof(XvPortRec)*XvplxNumPorts[na]);
      if (!pps)
	{
	  FatalError("XvplxInitAdaptors: coudn't alloc Port struct\n");
	}
      pp = pps;

      for (np = 0; np<XvplxNumPorts[na]; np++)
	{
	  int pno;
	  id = FakeClientID(0);
	  
	  if (!AddResource(id, XvRTPort, pp))
	    {
	      FatalError("XvddInitScreen: couldn't add Port resource\n");
	    }
	  
	  /* MANUALLY FILL IN PORT STRUCTURE */
	  
	  pp->id = id;
	  pp->pAdaptor = pa;
	  pp->pDraw = (DrawablePtr)NULL;
	  pp->grab.client = (ClientPtr)NULL;
	  pp->time = currentTime;
	  pp->pNotify = (XvPortNotifyPtr)NULL;
	  ppp = (XvplxPortPtr)xalloc(sizeof(XvplxPortRec));
	  pp->devPriv.ptr = (pointer)ppp;

	  ppp->pGC = (GCPtr)NULL;
	  ppp->pEncoding = pes; /* XXX---SHOULD BE A COMMAND LINE ARG */

	  if(na == 0) pno = PORT_D;
	  else if (np == 0) pno = input1_port;
	  else pno = output0_port;
	  ppp->hue = ((plxvideoport[pno].knob.hue * 2000) / 256) - 1000;
	  ppp->saturation = ((plxvideoport[pno].knob.saturation * 2000) / 256) -
				1000;
	  ppp->brightness = ((plxvideoport[pno].knob.bright * 2000) / 256) -
				1000;
	  ppp->contrast = ((plxvideoport[pno].knob.contrast * 2000) / 256) -
				1000;
	  ppp->serialNumber = 0;
	  ppp->pDrawStill = (DrawablePtr)NULL;
	  ppp->channel = na + np + 1;
	  if(ppp->channel == 3) ppp->channel = PLX_OUTPUT_0;

	  /* SAVE HANDLE TO DEVICE DRIVER */

	  /* SET DEFAULT ENCODING; NOT GRAY */

	  ppp->gray = FALSE;

	  pp++;
	  
	}

      nf = 0;
      tf = XvplxNumFormats[na];
      if(na && (xv->opt.board_code == OPT_2))
	--tf;
      pfs = (XvFormatPtr)xalloc(sizeof(XvFormatRec)*tf);
      if (!pfs)
	{
	  FatalError("XvplxInitAdaptors: couldn't alloc Format struct\n");
	}
      pf = pfs;

      pd = pScreen->allowedDepths;
      for (nd=0; nd<pScreen->numDepths; nd++)
	{
	  if (pd->depth >= 24)
	    {
	      pv = pd->vids;
	      for (nv=0; nv<pd->numVids; nv++)
		{
		  if (nf >= tf) 
		    {
		      tf *= 2;
		      pfs = (XvFormatPtr)xrealloc(pfs, sizeof(XvFormatRec)*tf);
		      pf = &pfs[nf];
		    }
		  
		  pf->depth = pd->depth;
		  pf->visual = *pv;
		  pf++;
		  nf++;
		  pv++;
		}
	    }
	  pd++;
	}

      /* FIND STATIC GRAY VISUAL if not rgb card */

      if(tf > 1)
      {
	pvis = pScreen->visuals;
	for (nvis=0; nvis<pScreen->numVisuals; nvis++)
	{
	  if ((pvis->class == StaticGray) && (pvis->nplanes == 8)) break;
	  pvis++;
	}

      /* IF WE FOUND ONE THEN FIND MATCHING DEPTH VISUALID */

	if (nvis < pScreen->numVisuals)
	{
	  pd = pScreen->allowedDepths;
	  for (nd=0; nd<pScreen->numDepths; nd++)
	    {
	      if (pd->depth == 8)
		{
		  pv = pd->vids;
		  for (nv=0; nv<pd->numVids; nv++)
		    {
		      if (pvis->vid == *pv)
			{
			  if (nf >= tf) 
			    {
			      tf *= 2;
			      pfs = (XvFormatPtr)
				xrealloc(pfs, sizeof(XvFormatRec)*tf);
			      pf = &pfs[nf];
			    }
			  pf->depth = pd->depth;
			  pf->visual = *pv;
			  pf++;
			  nf++;
			  break;
			}
		      pv++;
		    }
		}
	      pd++;
	    }
	}
      }

      XvplxNumFormats[na] = nf;

      pp = pps; /*+na; */
      pa->base_id = pp->id;
      pa->type = (na == 0) ? XvInputMask : XvInputMask | XvOutputMask;
      if(na == 0)
	pa->name = "Parallax Xvideo Base";
      else
	pa->name = "Parallax Xvideo Option";
      pa->nEncodings = ne;
      pa->pEncodings = pes;
      pa->nPorts = np;
      pa->pPorts = pps;
      pa->nFormats = nf;
      pa->pFormats = pfs;
      pa->pScreen = pScreen;
      pa->ddAllocatePort = XvplxAllocatePort;
      pa->ddFreePort = XvplxFreePort;
      pa->ddPutVideo = XvplxPutVideo;
      pa->ddPutStill = XvplxPutStill;
      pa->ddGetVideo = XvplxGetVideo;
      pa->ddGetStill = XvplxGetStill;
      pa->ddStopVideo = XvplxStopVideo;
      pa->ddSetPortAttribute = XvplxSetPortAttribute;
      pa->ddGetPortAttribute = XvplxGetPortAttribute;
      pa->ddQueryBestSize = XvplxQueryBestSize;

      /* ALLOWS QUICK ACCESS TO PORTS BY PLX DD MODULE. */

      XvplxPorts[pScreen->myNum][na] = pps;
      ++pa;
    }

  pxvs->nAdaptors = na;
  pxvs->pAdaptors = pas;

  /* ALLOCATE ATOMS FOR PORT ATTRIBUTES --- MORE LATER */

  XvplxEncoding = MakeAtom("XV_ENCODING", strlen("XV_ENCODING"), xTrue);
  XvplxHue = MakeAtom("XV_HUE", strlen("XV_HUE"), xTrue);
  XvplxSaturation = MakeAtom("XV_SATURATION", strlen("XV_SATURATION"), xTrue);
  XvplxBrightness = MakeAtom("XV_BRIGHTNESS", strlen("XV_BRIGHTNESS"), xTrue);
  XvplxContrast = MakeAtom("XV_CONTRAST", strlen("XV_CONTRAST"), xTrue);
  XvplxValueRecall = MakeAtom("XV_VALUE_RECALL",strlen("XV_VALUE_RECALL"),xTrue);
  XvplxValueSave = MakeAtom("XV_VALUE_SAVE",strlen("XV_VALUE_SAVE"),xTrue);
  XvplxCrop = MakeAtom("XV_CROP",strlen("XV_CROP"),xTrue);
  XvplxParameterString = MakeAtom("XV_VIDEO_PARAMETER_STRING",
			strlen("XV_VIDEO_PARAMETER_STRING"),xTrue);

  return Success;

}


static Bool
XvplxCloseScreen(ii, pScreen)
int ii;
ScreenPtr pScreen;

{
  int np,na,j;
  XvAdaptorPtr pa;
  XvPortPtr pp;
  XvScreenPtr pxvs;
  XvplxScreenPtr props;

  plxFreeResources(pScreen); /* Free jpeg buffer */
  pxvs = (XvScreenPtr) pScreen->devPrivates[XvScreenIndex].ptr;
  props = (XvplxScreenPtr)pxvs->devPriv.ptr;


  if (!pxvs) return TRUE;
#ifdef NO_PLX_VIDEO
  pScreen->WindowExposures = props->WindowExposures;
#endif

  pa = pxvs->pAdaptors;

  for (na=0; na<pxvs->nAdaptors; na++, pa++)
    {
      if (pa->pFormats) xfree(pa->pFormats);
      if (pa->pEncodings) xfree(pa->pEncodings);
      pp = pa->pPorts;
      if (pa->pPorts) xfree(pa->pPorts);
    }

  xfree(pxvs->pAdaptors);
  pxvs->pAdaptors = (XvAdaptorPtr)NULL;
  xvplx_num_adaptors = 1;	/* reset to avoid overflow */

  return TRUE;
}

static int
XvplxAllocatePort(port, pPort, ppPort)

unsigned long port;
XvPortPtr pPort;
XvPortPtr *ppPort;

{
  *ppPort = pPort;
  return Success;
}

static int
XvplxFreePort(pPort)

XvPortPtr pPort;

{
  return Success;
}

static int
XvplxQueryAdaptors(pScreen, p_nAdaptors, p_pAdaptors)

ScreenPtr pScreen;
XvAdaptorPtr *p_pAdaptors;
int *p_nAdaptors;

{

  XvScreenPtr pxvs;

  pxvs = (XvScreenPtr) pScreen->devPrivates[XvScreenIndex].ptr;

  *p_nAdaptors = pxvs->nAdaptors;
  *p_pAdaptors = pxvs->pAdaptors;

  return (Success);

}

static void
XvplxConstrainVideo(pWin, pe,
		    vid_x, vid_y, vid_w, vid_h,
		    drw_x, drw_y, drw_w, drw_h,
		    p_vid_x, p_vid_y, p_vid_w, p_vid_h,
		    p_drw_x, p_drw_y, p_drw_w, p_drw_h)

WindowPtr pWin;
XvEncodingPtr pe;
INT16 vid_x, vid_y, drw_x, drw_y;
CARD16 vid_w, vid_h, drw_w, drw_h;
INT16 *p_vid_x, *p_vid_y, *p_drw_x, *p_drw_y;
CARD16 *p_vid_w, *p_vid_h, *p_drw_w, *p_drw_h;

{
  ScreenPtr pScreen;
  int u,v;

  /* NOTE WE ARE USING SIGNED VALUES FOR EXTENTS; THE CALCULATIONS BELOW COULD
     PRODUCE NEGATIVE VALUES WHICH SHOULD BE INTERPRETED AS A NULL VIDEO
     REGION */

  /* INITIALIZE RESULTS TO ZERO IN CASE WE ENCOUNTER A ZERO EXTENT SOMEWHERE
     ALONG THE WAY */

  /* USE A LOOP TO HANDLE SIZE ERROR */

  for (;;)
    {
      
      /* FIRST SEE IF VID X,Y ARE OUT OF BOUNDS; NULL REGION RESULTS */
      
      if (vid_x >= (INT16 )pe->width || vid_y >= (INT16 )pe->height || 
	  vid_w <= 0 || vid_h <= 0) break;
      
      /* CLIP LOWER BOUNDS OF SOURCE REGION */
      
      if (vid_x < 0)
	{
	  u = (drw_w*(vid_w + vid_x))/vid_w;
	  drw_x += drw_w - u;
	  drw_w = u;
	  vid_w += vid_x;
	  vid_x = 0;
	  if (drw_w <= 0 || vid_w <= 0) break;
	}
      
      if (vid_y < 0)
	{
	  u = (drw_h*(vid_h + vid_y))/vid_h;
	  drw_y += drw_h - u;
	  drw_h = u;
	  vid_h += vid_y;
	  vid_y = 0;
	  if (drw_h <= 0 || vid_h <= 0) break;
	}
      
      /* CLIP UPPER BOUNDS OF SOURCE REGION */
      
      if (vid_x + vid_w > pe->width)
	{
	  drw_w = (drw_w*(pe->width - vid_x))/vid_w;
	  vid_w = pe->width - vid_x;
	  if (drw_w <= 0 || vid_w <= 0) break;
	}
      
      if (vid_y + vid_h > pe->height)
	{
	  drw_h = (drw_h*(pe->height - vid_y))/vid_h;
	  vid_h = pe->height - vid_y;
	  if (drw_h <= 0 || vid_h <= 0) break;
	}
      
      /* CLIP THE DST AREA TO THE LOWER BOUNDS OF SCREEN */
      
      if (pWin->drawable.x + drw_x < 0)
	{
	  u = drw_w + (pWin->drawable.x + drw_x);
	  v = (vid_w*u)/drw_w;
	  vid_x += vid_w - v;
	  vid_w = v;
	  drw_x = -pWin->drawable.x;
	  drw_w = u;
	  if (drw_w <= 0 || vid_w <= 0) break;
	}
      
      if (pWin->drawable.y + drw_y < 0)
	{
	  if (drw_h <= 0) break;
	  u = drw_h + (pWin->drawable.y + drw_y);
	  v = (vid_h*u)/drw_h;
	  vid_y += vid_h - v;
	  vid_h = v;
	  drw_y = -pWin->drawable.y;
	  drw_h = u;
	  if (drw_h <= 0 || vid_h <= 0) break;
	}
      
      /* CLIP THE DST AREA TO THE UPPER BOUNDS OF SCREEN */
      
      pScreen = pWin->drawable.pScreen;
      
      if (pWin->drawable.x + drw_x + drw_w > pScreen->width)
	{
	  u = pScreen->width - pWin->drawable.x - drw_x;
	  vid_w = (vid_w*u)/drw_w;
	  drw_w = u;
	  if (drw_w <= 0 || vid_w <= 0) break;
	}
      
      if (pWin->drawable.y + drw_y + drw_h > pScreen->height)
	{
	  u = pScreen->height - pWin->drawable.y - drw_y;
	  vid_h = (vid_h*u)/drw_h;
	  drw_h = u;
	  if (drw_h <= 0 || vid_h <= 0) break;
	}
      
      /* CHECK DESTINATION SIZE AGAINST SOURCE VIDEO SIZE */
      
      if (drw_w > vid_w)
	{
	  drw_h = (drw_h * vid_w)/drw_w;
	  drw_w = vid_w;
	  if (drw_w <= 0 || vid_w <= 0) break;
	}
      
      if (drw_h > vid_h)
	{
	  drw_w = (drw_w * vid_h)/drw_h;
	  drw_h = vid_h;
	  if (drw_h <= 0 || vid_h <= 0) break;
	}
      
      *p_vid_x = vid_x;
      *p_vid_y = vid_y;
      *p_vid_w = vid_w;
      *p_vid_h = vid_h;
      *p_drw_x = drw_x;
      *p_drw_y = drw_y;
      *p_drw_w = drw_w;
      *p_drw_h = drw_h;

      return;

    }

  *p_vid_x = 0;
  *p_vid_y = 0;
  *p_vid_w = 0;
  *p_vid_h = 0;
  *p_drw_x = 0;
  *p_drw_y = 0;
  *p_drw_w = 0;
  *p_drw_h = 0;

}

static int
XvplxPutVideo(client, pDraw, pPort, pGC, vid_x, vid_y, vid_w, vid_h,
	       drw_x, drw_y, drw_w, drw_h)

ClientPtr client;
DrawablePtr pDraw;
XvPortPtr pPort;
GCPtr pGC;
INT16 vid_x, vid_y, drw_x, drw_y;
CARD16 vid_w, vid_h, drw_w, drw_h;

{
  XvplxPortPtr ppp;
  GCPtr pGCTemp;
  int vis;
  BoxPtr pBox;
  RegionPtr pClip;
  xVideoArgs xva;
  int i, port; 
  XvEncodingPtr pe;
  TV2Priv *tv2_priv;

  if (pDraw->type != DRAWABLE_WINDOW)
    {
      pPort->pDraw = (DrawablePtr)NULL;
      return BadAlloc;
    }

  plxSetWindowType(pDraw,XV_WINDOW_TYPE);
  ppp = (XvplxPortPtr)pPort->devPriv.ptr;
  pe = ppp->pEncoding;

  if(ppp->channel == PLX_UNKNOWN)
    return(BadMatch);

  if(ppp->channel == PLX_OUTPUT_0)
    ppp->channel = PLX_INPUT_1;

  ppp->pGC = pGC;

  ppp->vx = vid_x;
  ppp->vy = vid_y;
  ppp->vw = vid_w;
  ppp->vh = vid_h;
  ppp->dx = drw_x;
  ppp->dy = drw_y;
  ppp->dw = drw_w;
  ppp->dh = drw_h;

  /* CONSTRAIN THE REGIONS TO THOSE THAT ARE SUPPORTED; MAINTAIN ASPECT
     RATIO WHEN POSSIBLE */

  XvplxConstrainVideo(pDraw, pe, 
		      vid_x, vid_y, vid_w, vid_h,
		      drw_x, drw_y, drw_w, drw_h,
		      &vid_x, &vid_y, &vid_w, &vid_h,
		      &drw_x, &drw_y, &drw_w, &drw_h);


  ppp->Dx = pDraw->x;
  ppp->Dy = pDraw->y;

  xva.vx = ppp->cvx = vid_x;
  xva.vy = ppp->cvy = vid_y;
  xva.vw = ppp->cvw = vid_w;
  xva.vh = ppp->cvh = vid_h;
  xva.x = ppp->cdx = drw_x;
  xva.y = ppp->cdy = drw_y;
  xva.w = ppp->cdw = drw_w;
  xva.h = ppp->cdh = drw_h;

  plxv_switch_context(xv_ctx);

  port = (ppp->channel == PLX_INPUT_0) ? PORT_D : input1_port;

  if (pDraw->depth == 8)
  {
    if(ppp->gray == FALSE)
    {
      plxv_input_config(port, "BW");
      ppp->gray = TRUE;
    }
  }
  else if(ppp->gray)
  {
    ppp->gray = FALSE;
    plxv_input_config(port, "COLOR");
  }

  tv2_priv = XvplxSetupGC(pGC,pDraw,pPort);
  xva.vy += ((tv2_priv->standard == PLX_NTSC) ? NTSC_BLANK : PAL_BLANK);
  if( ppp->Crop)
  i = PlxVideoProc(pDraw, pGC, X_VideoLive, &xva, client);
  else
  i = PlxVideoProc(pDraw, pGC, X_VideoSqueezeLive, &xva, client);


  pPort->pDraw = pDraw;

  return (Success);
}

static int
XvplxPutStill(client, pDraw, pPort, pGC, vid_x, vid_y, vid_w, vid_h,
	       drw_x, drw_y, drw_w, drw_h)
ClientPtr client;
DrawablePtr pDraw;
XvPortPtr pPort;
GCPtr pGC;
INT16 vid_x, vid_y, drw_x, drw_y;
{
  XvplxPortPtr ppp;
  GCPtr pGCTemp;
  int vis,nBox,x1,x2,y1,y2;
  xVideoArgs xva;
  BoxPtr pBox;
  RegionPtr pClip;
  Bool src_area_changed = FALSE;
  Bool dst_area_changed = FALSE;
  int i, port;
  XvEncodingPtr pe;
  TV2Priv *tv2_priv;

  if (pDraw->type != DRAWABLE_WINDOW)
    {
      pPort->pDraw = (DrawablePtr)NULL;
      return BadAlloc;
    }
  plxSetWindowType(pDraw,XV_WINDOW_TYPE);

  ppp = (XvplxPortPtr)pPort->devPriv.ptr;
  pe = ppp->pEncoding;

  if(ppp->channel == PLX_UNKNOWN)
    return(BadMatch);

  if(ppp->channel == PLX_OUTPUT_0)
    ppp->channel = PLX_INPUT_1;

  if(pPort->pDraw == NULL)
  {
    ppp->vx = vid_x;
    ppp->vy = vid_y;
    ppp->vw = vid_w;
    ppp->vh = vid_h;
    ppp->dx = drw_x;
    ppp->dy = drw_y;
    ppp->dw = drw_w;
    ppp->dh = drw_h;
  }

  /* CONSTRAIN THE REGIONS TO THOSE THAT ARE SUPPORTED; MAINTAIN ASPECT
     RATIO WHEN POSSIBLE */

  XvplxConstrainVideo(pDraw, pe, 
		      vid_x, vid_y, vid_w, vid_h,
		      drw_x, drw_y, drw_w, drw_h,
		      &vid_x, &vid_y, &vid_w, &vid_h,
		      &drw_x, &drw_y, &drw_w, &drw_h);


  if(pPort->pDraw == NULL)
  {
    ppp->Dx = pDraw->x;
    ppp->Dy = pDraw->y;

    xva.vx = ppp->cvx = vid_x;
    xva.vy = ppp->cvy = vid_y;
    xva.vw = ppp->cvw = vid_w;
    xva.vh = ppp->cvh = vid_h;
    xva.x = ppp->cdx = drw_x;
    xva.y = ppp->cdy = drw_y;
    xva.w = ppp->cdw = drw_w;
    xva.h = ppp->cdh = drw_h;
  }
  else
  {
    xva.vx = vid_x;
    xva.vy = vid_y;
    xva.vw = vid_w;
    xva.vh = vid_h;
    xva.x = drw_x;
    xva.y = drw_y;
    xva.w = drw_w;
    xva.h = drw_h;
  }

  plxv_switch_context(xv_ctx);

  port = (ppp->channel == PLX_INPUT_0) ? PORT_D : input1_port;

  if (pDraw->depth == 8)
  {
    if(ppp->gray == FALSE)
    {
      plxv_input_config(port, "BW");
      ppp->gray = TRUE;
    }
  }
  else if(ppp->gray)
  {
    ppp->gray = FALSE;
    plxv_input_config(port, "COLOR");
  }

  tv2_priv = XvplxSetupGC(pGC,pDraw,pPort);

  xva.vy += ((tv2_priv->standard == PLX_NTSC) ? NTSC_BLANK : PAL_BLANK);
  i = PlxVideoProc(pDraw, pGC, X_VideoSqueezeStill, &xva, client);

  if(pPort->pDraw)
  {
    xva.vx = ppp->cvx;
    xva.vy = ppp->cvy;
    xva.vw = ppp->cvw;
    xva.vh = ppp->cvh;
    xva.x = ppp->cdx;
    xva.y = ppp->cdy;
    xva.w = ppp->cdw;
    xva.h = ppp->cdh;
    MyUSleep(33333);
    xva.vy += ((tv2_priv->standard == PLX_NTSC) ? NTSC_BLANK : PAL_BLANK);
    PlxVideoProc(pPort->pDraw, ppp->pGC, X_VideoSqueezeLive, &xva, client);
  }
/*
  else
    pPort->pDraw = pDraw;
*/
  return (Success);
}


static int
XvplxGetVideo(client, pDraw, pPort, pGC, vid_x, vid_y, vid_w, vid_h,
	      drw_x, drw_y, drw_w, drw_h)
ClientPtr client;
DrawablePtr pDraw;
XvPortPtr pPort;
GCPtr pGC;
INT16 vid_x, vid_y, drw_x, drw_y;
CARD16 vid_w, vid_h, drw_w, drw_h;
{
  XvplxPortPtr ppp;
  GCPtr pGCTemp;
  int vis,nBox,x1,x2,y1,y2;
  BoxPtr pBox;
  RegionPtr pClip;
  Bool src_area_changed = FALSE;
  Bool dst_area_changed = FALSE;
  xVideoArgs xva;
  int i;
  XvEncodingPtr pe;
  TV2Priv *tv2_priv;

  if (pDraw->type != DRAWABLE_WINDOW)
    {
      pPort->pDraw = (DrawablePtr)NULL;
      return BadAlloc;
    }
  plxSetWindowType(pDraw,XV_WINDOW_TYPE);

  ppp = (XvplxPortPtr)pPort->devPriv.ptr;
  pe = ppp->pEncoding;

  if(ppp->channel == PLX_INPUT_0)
	return(BadMatch);

  ppp->channel = PLX_OUTPUT_0;

  ppp->vx = vid_x;
  ppp->vy = vid_y;
  ppp->vw = vid_w;
  ppp->vh = vid_h;
  ppp->dx = drw_x;
  ppp->dy = drw_y;
  ppp->dw = drw_w;
  ppp->dh = drw_h;

  /* CONSTRAIN THE REGIONS TO THOSE THAT ARE SUPPORTED; MAINTAIN ASPECT
     RATIO WHEN POSSIBLE */

  XvplxConstrainVideo(pDraw, pe, 
		      vid_x, vid_y, vid_w, vid_h,
		      drw_x, drw_y, drw_w, drw_h,
		      &vid_x, &vid_y, &vid_w, &vid_h,
		      &drw_x, &drw_y, &drw_w, &drw_h);


  ppp->Dx = pDraw->x;
  ppp->Dy = pDraw->y;

  xva.vx = ppp->cvx = vid_x;
  xva.vy = ppp->cvy = vid_y;
  xva.vw = ppp->cvw = vid_w;
  xva.vh = ppp->cvh = vid_h;
  xva.x = ppp->cdx = drw_x;
  xva.y = ppp->cdy = drw_y;
  xva.w = ppp->cdw = drw_w;
  xva.h = ppp->cdh = drw_h;

  plxv_switch_context(xv_ctx);

  tv2_priv =XvplxSetupGC(pGC,pDraw,pPort);
  xva.vy += ((tv2_priv->standard == PLX_NTSC) ? NTSC_BLANK : PAL_BLANK);
  i = PlxVideoProc(pDraw, pGC, X_VideoScaleOutput, &xva, client);

  pPort->pDraw = pDraw;

  return (Success);
}


static int
XvplxGetStill(client, pDraw, pPort, pGC, vid_x, vid_y, vid_w, vid_h,
	      drw_x, drw_y, drw_w, drw_h)
ClientPtr client;
DrawablePtr pDraw;
XvPortPtr pPort;
GCPtr pGC;
INT16 vid_x, vid_y, drw_x, drw_y;
CARD16 vid_w, vid_h, drw_w, drw_h;

{
  return BadMatch;
}


static int
XvplxStopVideo(client, pPort, pDraw)
ClientPtr client;
XvPortPtr pPort;
DrawablePtr pDraw;

{
  XvplxPortPtr ppp;
  GCPtr pGCTemp;
  xVideoSelectArgs args;
  XvEncodingPtr pe;
  int i, port;

  ppp = (XvplxPortPtr)pPort->devPriv.ptr;
  pe = ppp->pEncoding;

  plxv_switch_context(xv_ctx);

  if(ppp->channel != PLX_OUTPUT_0)
  {
    i = plxVideoStop(pDraw);
    if(ppp->gray)
    {
      port = (ppp->channel == PLX_INPUT_0) ? PORT_D : input1_port;
      MyUSleep(33333);
      plxv_input_config(port, "COLOR");
      ppp->gray = 0;
    }
  }
  else if(pPort->pDraw)
  {
    args.channel = PLX_IO_OFF;
    args.standard = pe->id >> 16;
    args.format = pe->id & 0xffff;
    args.type = PLX_RGB24;
    i = plxVideoOutputSelect(pPort->pDraw, ppp->pGC, &args);
  }
  pPort->pDraw = NULL;

  return (Success);
}


static int
XvplxSetPortAttribute(client, pPort, attribute, value)
ClientPtr client;
XvPortPtr pPort;
Atom attribute;
INT32 value;

{
  int ii;
  unsigned char lut[256];
  XvAdaptorPtr pa;
  XvEncodingPtr pe;
  XvplxPortPtr ppp;
  float contrast, brightness;
  int on_off;
  int rop_value;
  int valsetflag;
  xVidValueArgs xva;
  int i, port;

  ppp = (XvplxPortPtr)pPort->devPriv.ptr;
  pa = pPort->pAdaptor;
  pe = pa->pEncodings;
  valsetflag = 0;
  plxv_switch_context(xv_ctx);
  if (ppp->channel == PLX_INPUT_0)
        port = PORT_D;
  else if (ppp->channel == PLX_INPUT_1)
        port = input1_port;
  else
        port = output0_port;

  /* FIND THE ATTRIBUTE AND SET IT; XXX --- MORE LATER */

  if (attribute == XvplxEncoding)
    {
      for (ii=0; ii<pa->nEncodings; ii++)
	{
	  if (pe->id == (XvEncodingID)value)
	    {

	      /* SAVE NEW ENCODING ID */

	      ppp->pEncoding = pe;

	      break;
	    }
	  pe++;
	}
      if (ii>=pa->nEncodings) return _XvBadEncoding;
    }
  else if (attribute == XvplxHue)
    {

      if (value < -1000) value = -1000;
      if (value > 1000) value = 1000;

      /* MAP -1000..1000 into 0..255 */

      rop_value = (255*(value+1000))/2000;

      plxvideoport[port].knob.hue = rop_value;
      ppp->hue = value;
      valsetflag = 1;
    }
  else if (attribute == XvplxSaturation)
    {
      if (value < -1000) value = -1000;
      if (value > 1000) value = 1000;

      /* MAP -1000..1000 into 0..255 */

      rop_value = (255*(value+1000))/2000;
      plxvideoport[port].knob.saturation = rop_value;

      ppp->saturation = value;
      valsetflag = 1;
    }
  else if (attribute == XvplxBrightness)
    {
      if (value < -1000) value = -1000;
      if (value > 1000) value = 1000;

      /* MAP -1000..1000 into 0..255 */

      rop_value = (255*(value+1000))/2000;
      plxvideoport[port].knob.bright = rop_value;

      ppp->brightness = value;
      valsetflag = 1;
    }
  else if (attribute == XvplxContrast)
    {
      if (value < -1000) value = -1000;
      if (value > 1000) value = 1000;

      /* MAP -1000..1000 into 0..255 */

      rop_value = (255*(value+1000))/2000;

      plxvideoport[port].knob.contrast = rop_value;
      ppp->contrast = value;
      valsetflag = 1;
    }
  else if(attribute == XvplxCrop)
  {
      ppp->Crop= (Bool)value;
  }
  else if(attribute == XvplxValueRecall) 
  {
  /* Load NVRAM from table selected by "value" */
     plxVideoValueRecall (value);
  }
  else if(attribute == XvplxValueSave)
  {
   /* Store given data into NVRAM selected by "value" */
      plxVideoValueSave (value);
  }
  else if(attribute == XvplxParameterString)
  {
       switch(value)
       {
	case InvertedStringID:
	     ppp->Inverted=TRUE;
	     break;
	case NonInvertedStringID:
	     ppp->Inverted=FALSE;
	     break;
	case BWStringID:
	     ppp->gray=TRUE;
	     break;
	case ColorStringID:
	     ppp->gray=FALSE;
	     break;
	case GenlockStringID:
	     ppp->Genlock=TRUE;
	     break;
	case FreerunStringID:
	     ppp->Genlock=FALSE;
	     break;
       }
  }
  else return BadMatch;
  if(valsetflag)
  {
    plxv_set_knobs(port, &plxvideoport[port].knob);
  }

  XvdiSendPortNotify(pPort, attribute, value);

  return (Success);
}

static int
XvplxGetPortAttribute(client, pPort, attribute, p_value)
ClientPtr client;
XvPortPtr pPort;
Atom attribute;
INT32 *p_value;

{

  XvplxPortPtr ppp;
  ppp = (XvplxPortPtr)pPort->devPriv.ptr;

  /* FIND THE ATTRIBUTE AND RETURN ITS VALUE; XXX --- MORE LATER */

  if (attribute == XvplxEncoding)
    {
      *p_value = ppp->pEncoding->id;
    }
  else if (attribute == XvplxHue)
    {
      *p_value = ppp->hue;
    }
  else if (attribute == XvplxSaturation)
    {
      *p_value = ppp->saturation;
    }
  else if (attribute == XvplxBrightness)
    {
      *p_value = ppp->brightness;
    }
  else if (attribute == XvplxContrast)
    {
      *p_value = ppp->contrast;
    }
  else 
      return BadMatch;

  return Success;

}

static int
XvplxQueryBestSize(client, pPort, motion, vid_w, vid_h, drw_w, drw_h, p_w, p_h)
ClientPtr client;
XvPortPtr pPort;
CARD8 motion;
CARD16 vid_w,vid_h;
CARD16 drw_w,drw_h;
CARD16 *p_w, *p_h;

{
  XvplxPortPtr ppp;
  XvEncodingPtr pe;
  int vid_x,vid_y,drw_x,drw_y;
  extern WindowPtr *WindowTable;

  vid_x = vid_y = drw_x = drw_y = 0;

  ppp = (XvplxPortPtr)pPort->devPriv.ptr;
  pe = ppp->pEncoding;

  /* USE THE SAME GEOMETRY CONSTRAINT PROC THAT WE USE FOR GET AND PUT */

  XvplxConstrainVideo(WindowTable[pPort->pAdaptor->pScreen->myNum], pe, 
		      vid_x, vid_y, vid_w, vid_h,
		      drw_x, drw_y, drw_w, drw_h,
		      &vid_x, &vid_y, &vid_w, &vid_h,
		      &drw_x, &drw_y, p_w, p_h);

  return Success;

}

void
XvplxWindowExposures(pWin, pReg, pOtherReg)
WindowPtr pWin;
RegionPtr pReg;
RegionPtr pOtherReg;

{

  ScreenPtr pScreen;
  XvScreenPtr pxvs;
  XvplxScreenPtr props;
  XvPortPtr pp;
  XvplxPortPtr ppp;
  int i;
  plxScreenPtr  pPlxPrivate;
  
  pScreen = pWin->drawable.pScreen;
  pPlxPrivate =  GetPlxPrivate(pScreen);

  pxvs = (XvScreenPtr)pWin->drawable.pScreen->devPrivates[XvScreenIndex].ptr;
  props = (XvplxScreenPtr)pxvs->devPriv.ptr;

#ifdef NO_PLX_VIDEO
  pScreen->WindowExposures = props->WindowExposures;
#endif

  for(i = 0; i < xvplx_num_adaptors; ++i)
  {
    pp = XvplxPorts[pWin->drawable.pScreen->myNum][i];

    if ((pp->pDraw) && (((WindowPtr)pp->pDraw)->realized))
    {
      if(plxrestore == 0)
      {
	ppp = (XvplxPortPtr)pp->devPriv.ptr;
	if (pp->pDraw->serialNumber != ppp->serialNumber)
	{
	     XvplxRestartVideo(pp);
	     pPlxPrivate->video_restored=1;
	}
      }
    }
  }
#ifdef NO_PLX_VIDEO

  (* pScreen->WindowExposures)(pWin, pReg, pOtherReg);

  pScreen->WindowExposures = XvplxWindowExposures;
#endif

}

XvplxRestartVideo(pPort)

XvPortPtr pPort;

{
  XvplxPortPtr ppp;
  
  ppp = (XvplxPortPtr)pPort->devPriv.ptr;

  if (ppp->pGC)
    {

      if (ppp->pGC->serialNumber != pPort->pDraw->serialNumber)
	ValidateGC(pPort->pDraw, ppp->pGC);

      XvplxPutVideo((ClientPtr)NULL, 
		    pPort->pDraw, pPort, ppp->pGC, 
		    ppp->vx, ppp->vy, ppp->vw, ppp->vh,
		    ppp->dx, ppp->dy, ppp->dw, ppp->dh);
    }
  else
    {
      XvdiSendVideoNotify(pPort, pPort->pDraw, XvPreempted);
    }

}
#endif /* USE_XV */
