/***********************************************************
Copyright 1991 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/*
** File: 
**
**   xvplx.c --- Xv Parallax device dependent module.
**
** Author: 
**
**   David Carver (Digital Workstation Engineering/Project Athena)
**
** Revisions:
**
**   26.02.92 Dave Arnold
**     - Initiall Parallax port
**
**   26.11.91 Carver
**     - optimized PutStill to not redraw enable plane between like stills.
**     - fixed multi-screen support
**     - added new call to pip_installed()
**
**   18.09.91 Carver
**     - changed interface to libpip.c; now uses pip_src_area and pip_dst_area
**       instead of pip_source_area, pip_origin, and pip_size.
**     - fixed constrain procedure to clip video destination to screen bounds
**
**   10.09.91 Carver
**     - fixed calls to fill_video_enable some were overrunning bounds
**       of screen.
**
**   29.08.91 Carver
**     - change: unrealizing windows no longer preempts video
**     - added support for duty cycle (video priority) management
**     - added support for video in StaticGray windows
**     - added include of cfb.h; needed to access 8 plane gc priv structure
**     - changed interfaces to rop.c and libpip.c to use RopPtr uniformly
**     - changed ultrix conditional compile so this file works for both
**       sun and dec boards
**
**   26.06.91 Carver
**     - fixed GC wrapping logic in a big way
**     - fixed XvropValidateGC; it had the wrong argument definition
**       can't believe that it ever worked.
**     - fixed XvropStopVideo to check to see if the GC still exists
**       before it references a pointer to it.
**
**   11.06.91 Carver
**     - changed SetPortControl to SetPortAttribute
**     - changed GetPortControl to GetPortAttribute
**     - changed QueryBestSize
**
**   04.06.91 Carver
**     - completed implementing port controls
**     - used code from R. Ulichney to to contrast and brightness calc's
**     - changed interface to libpip.c, use new libpip.h header file
**
**   31.05.91 Carver
**     - made big fixes to occlusion stuff
**     - fixed video image offset and size
**
**   23.05.91 Carver
**     - check for pip existance before initializing the adaptor
**
**   15.05.91 Carver
**     - version 2.0 upgrade
**
**   24.04.91 Carver
**     - updated to use 24 plane support.
**
**   19.03.91 Carver
**     - origional version coded to v1r4.
**
*/

#include <stdio.h>
#include <sys/types.h>

#define PSZ	32
#include "X.h"
#include "Xproto.h"
#include "scrnintstr.h"
#include "validate.h"
#include "windowstr.h"
#include "pixmapstr.h"
#include "gcstruct.h"
#include "dixstruct.h"
#include "cfb.h"                


#include "Xv.h"
#include "Xvproto.h"
#include "xvdix.h"
#include "tv2reg.h"
#include "plxInit.h"

#define M8VOFF 0x40000
#define M8SIZE (1152*900)
#define M24VOFF 0x140000
#define M24SIZE (1152*900*4)

#define GLOBAL
#include "xvplx.h"
#undef GLOBAL

#include "plxv.h"
#include "plxv_xpoint.h"
#include "xvideo.h"

#define NTSC_BLANK      45
#define PAL_BLANK       50
#define InvertedStringID	0
#define NonInvertedStringID	1
#define BWStringID		2
#define ColorStringID		3
#define GenlockStringID		4
#define FreerunStringID		5
#define XV_WINDOW_TYPE		2

typedef struct _PLXPort {
    int         io_type;
    XV_knob_rec knob;
    } PLXPortRec;

typedef struct _VidValueArgs {
	INT16 select B16;
	INT16 sat B16;
	INT16 cont B16;
	INT16 hue B16;
	INT16 brit B16;
	INT16 pad0 B16;
	INT32 pad1 B32;
} xVidValueArgs;
#define sz_xVidValueArgs 16

typedef struct _VideoSelectArgs {
	INT16 channel B16;
	INT16 standard B16;
	INT16 format B16;
	INT16 type B16;
} xVideoSelectArgs;
#define sz_xVideoSelectArgs 8;

typedef struct _VideoArgs {
	INT16 vx B16, vy B16, vw B16, vh B16;	/* input video area */
	INT16 x B16, y B16, w B16, h B16;	/* windowing cord's */
} xVideoArgs;
#define sz_xVideoArgs 16

extern PLXPortRec plxvideoport[4];

extern int input1_port;
extern int output0_port;
int plxrestore;
int i;


/* INTERNAL */
#define	X_VideoLive		12
#define X_VideoSqueezeLive	41
#define X_VideoSqueezeStill	42

#define X_VideoFlipLive		51
#define X_VideoFlipStill	52
#define X_VideoFlipScale	53
#define X_VideoFlipSqueezeLive	54
#define X_VideoFlipSqueezeStill	55
#define X_VideoScaleOutput	101
#define X_VideoOutputSelect	104

static Bool XvplxCloseScreen();
static int XvplxQueryAdaptors();
static int XvplxInitAdaptor();
static int XvplxAllocatePort();
static int XvplxFreePort();
static int XvplxPutVideo();
static int XvplxPutStill();
static int XvplxGetVideo();
static int XvplxGetStill();
static int XvplxStopVideo();
static int XvplxSetPortAttribute();
static int XvplxGetPortAttribute();
static int XvplxQueryBestSize();


static void XvplxClipNotify();
static void XvplxCopyWindow();
static Bool XvplxUnrealizeWindow();
void XvplxWindowExposures();

static void XvplxDestroyGC();
static void XvplxValidateGC();

static XvPortPtr XvplxPorts[MAXSCREENS][2];
XV_struct *xv_ctx;
extern XV_struct *PlxGetContext();
TV2Priv *init_priv_tv2();
static int tv2_priv_cnt=0;


TV2Priv *SetTv2Priv(DrawablePtr pDraw, GCPtr pGC)
{
    plxGCPtr	  privGC = GetPlxGCPrivate(pGC);
    plxPrivWinPtr plxWinPtr = plxGetPwinPriv(pDraw);
    TV2Priv* 	  tv2_priv;
    int  	  wid;

  if(privGC->video_priv==NULL)
  {
      tv2_priv = init_priv_tv2(pGC);
#ifdef NOTDEF
      tv2_priv->video_tag  = plxGetWid(pGC->pScreen,pDraw);
      tv2_priv->window_type=PLX_VIDEO;
#endif
      privGC->video_priv= (int *)tv2_priv;
  }
  else 
      tv2_priv= (TV2Priv *)privGC->video_priv;
  /* plxWinPtr->Video_Priv= (int *)tv2_priv; */
  return(tv2_priv);
}

TV2Priv *GetTv2Priv(GCPtr pGC)
{
int  i=0;
TV2Priv *tv2_priv;
plxGCPtr	privGC = GetPlxGCPrivate(pGC);

  if(privGC->video_priv)
	return((TV2Priv *)privGC->video_priv);
  return(NULL);
}

void XvplxVideoInit()
{
ScreenPtr 	pScreen;
ScreenInfo *pScreenInfo;
plxScreenPtr	pPlxPrivate;
int i_screen=0;
  
    pScreenInfo = &screenInfo;
    for ( i_screen = 0; i_screen < pScreenInfo->numScreens; i_screen++ )
    {
	    pScreen = pScreenInfo->screens[i_screen];
	    pPlxPrivate =  GetPlxPrivate(pScreen);
	    if(pPlxPrivate != NULL) 
 		{
		    xv_ctx = PlxGetContext(pScreen);
  			plxVideoInit(pScreen, xv_ctx);
		}
	}

}

TV2Priv *
init_priv_tv2(pGC)
GCPtr pGC;
{
TV2Priv *priv;

  priv = (TV2Priv *)xalloc(sizeof(TV2Priv));
  priv->window_type = PLX_GRAPHICS;
  priv->channel = 0;
  priv->standard = PLX_NTSC;
  priv->format = PLX_COMP;
  priv->type = PLX_RGB24;
  priv->gov_tag=0;
  priv->gov_mode=0;
  priv->xdirection = 0;
  priv->ydirection = 0;
  priv->video_tag = 0;
  priv->qtab = NULL;
  priv->htab = NULL;
  priv->pGC = pGC;
  priv->pScreen =  pGC->pScreen;
  priv->pWid = 0;
  return(priv);
}
